@@ -17,12 +17,15 @@ from web3.types import (
     Nonce,
     HexBytes,
 )
+from eth_utils import is_same_address
+from eth_typing import AnyAddress
 
 ETH_ADDRESS = "0x0000000000000000000000000000000000000000"
 
 logger = logging.getLogger(__name__)
 
 
+# TODO: Consider dropping support for ENS altogether and instead use AnyAddress
 AddressLike = Union[Address, ChecksumAddress, ENS]
 
 
@@ -48,7 +51,7 @@ def check_approval(method: Callable) -> Callable:
         need to be approved."""
 
     @functools.wraps(method)
-    def approved(self: Any, *args: Any) -> Any:
+    def approved(self: Any, *args: Any, **kwargs: Any) -> Any:
         # Check to see if the first token is actually ETH
         token = args[0] if args[0] != ETH_ADDRESS else None
         token_two = None
@@ -66,7 +69,7 @@ def check_approval(method: Callable) -> Callable:
             is_approved = self._is_approved(token_two)
             if not is_approved:
                 self.approve(token_two)
-        return method(self, *args)
+        return method(self, *args, **kwargs)
 
     return approved
 
@@ -263,9 +266,11 @@ class Uniswap:
     def erc20_contract(self, token_addr: AddressLike) -> Contract:
         return self._load_contract(abi_name="erc20", address=token_addr)
 
+    @functools.lru_cache()
     @supports([2])
-    def get_weth_address(self) -> Address:
-        address: Address = self.router.functions.WETH().call()
+    def get_weth_address(self) -> ChecksumAddress:
+        # Contract calls should always return checksummed addresses
+        address: ChecksumAddress = self.router.functions.WETH().call()
         return address
 
     def _load_contract(self, abi_name: str, address: AddressLike) -> Contract:
@@ -309,9 +314,16 @@ class Uniswap:
 
     @supports([2])
     def get_token_token_input_price(
-        self, token0: AddressLike, token1: AddressLike, qty: int
+        self, token0: AnyAddress, token1: AnyAddress, qty: int
     ) -> int:
         """Public price for token to token trades with an exact input."""
+        # If one of the tokens are WETH, delegate to appropriate call.
+        # See: https://github.com/shanefontaine/uniswap-python/issues/22
+        if is_same_address(token0, self.get_weth_address()):
+            return int(self.get_eth_token_input_price(token1, qty))
+        elif is_same_address(token1, self.get_weth_address()):
+            return int(self.get_token_eth_input_price(token0, qty))
+
         price: int = self.router.functions.getAmountsOut(
             qty, [token0, self.get_weth_address(), token1]
         ).call()[-1]
@@ -343,9 +355,17 @@ class Uniswap:
 
     @supports([2])
     def get_token_token_output_price(
-        self, token0: AddressLike, token1: AddressLike, qty: int
+        self, token0: AnyAddress, token1: AnyAddress, qty: int
     ) -> int:
         """Public price for token to token trades with an exact output."""
+        # If one of the tokens are WETH, delegate to appropriate call.
+        # See: https://github.com/shanefontaine/uniswap-python/issues/22
+        # TODO: Will these equality checks always work? (Address vs ChecksumAddress vs str)
+        if is_same_address(token0, self.get_weth_address()):
+            return int(self.get_eth_token_output_price(token1, qty))
+        elif is_same_address(token1, self.get_weth_address()):
+            return int(self.get_token_eth_output_price(token0, qty))
+
         price: int = self.router.functions.getAmountsIn(
             qty, [token0, self.get_weth_address(), token1]
         ).call()[0]
@@ -500,9 +520,8 @@ class Uniswap:
         """Convert tokens to ETH given an input amount."""
         # Balance check
         input_balance = self.get_token_balance(input_token)
-        cost = self.get_token_eth_input_price(input_token, qty)
-        if cost > input_balance:
-            raise InsufficientBalance(input_balance, cost)
+        if qty > input_balance:
+            raise InsufficientBalance(input_balance, qty)
 
         if self.version == 1:
             token_funcs = self.exchange_contract(input_token).functions
@@ -516,10 +535,14 @@ class Uniswap:
         else:
             if recipient is None:
                 recipient = self.address
+            amount_out_min = int(
+                (1 - self.max_slippage)
+                * self.get_token_eth_input_price(input_token, qty)
+            )
             return self._build_and_send_tx(
                 self.router.functions.swapExactTokensForETH(
                     qty,
-                    int((1 - self.max_slippage) * cost),
+                    amount_out_min,
                     [input_token, self.get_weth_address()],
                     recipient,
                     self._deadline(),
